import numba
import numpy as np
from math import sqrt, pow, cos, sin, asin

@numba.jit(nopython=True, fastmath=True)
def manhattan(array_x, array_y):
	n = array_x.shape[0]
	ret = 0.
	for i in range(n):
		ret += abs(array_x[i]-array_y[i])
	return ret

@numba.jit(nopython=True, fastmath=True)
def euclidean(array_x, array_y):
	n = array_x.shape[0]
	ret = 0.
	for i in range(n):
		ret += (array_x[i]-array_y[i])**2
	return sqrt(ret)

@numba.jit(nopython=True, fastmath=True)
def chebyshev(array_x, array_y):
	n = array_x.shape[0]
	ret = -1*np.inf
	for i in range(n):
		d = abs(array_x[i]-array_y[i])
		if d>ret: ret=d
	return ret

@numba.jit(nopython=True, fastmath=True)
def cosine(array_x, array_y):
	n = array_x.shape[0]
	xy_dot = 0.
	x_norm = 0.
	y_norm = 0.
	for i in range(n):
		xy_dot += array_x[i]*array_y[i]
		x_norm += array_x[i]*array_x[i]
		y_norm += array_y[i]*array_y[i]
	return 1.-xy_dot/(sqrt(x_norm)*sqrt(y_norm))

@numba.jit(nopython=True, fastmath=True)
def haversine(array_x, array_y):
	R = 6378.0
	radians = np.pi / 180.0
	lat_x = radians * array_x[0]
	lon_x = radians * array_x[1]
	lat_y = radians * array_y[0]
	lon_y = radians * array_y[1]
	dlon = lon_y - lon_x
	dlat = lat_y - lat_x
	a = (pow(sin(dlat/2.0), 2.0) + cos(lat_x) *
		cos(lat_y) * pow(sin(dlon/2.0), 2.0))
	return R * 2 * asin(sqrt(a))



import numpy as np
import numba
import hausdorff.distances as distances
from inspect import getmembers

def _find_available_functions(module_name):
	all_members = getmembers(module_name)
	available_functions = [member[0] for member in all_members 
						   if isinstance(member[1], numba.core.registry.CPUDispatcher)]
	return available_functions

@numba.jit(nopython=True, fastmath=True)
def _hausdorff(XA, XB, distance_function):
	nA = XA.shape[0]
	nB = XB.shape[0]
	cmax = 0.
	for i in range(nA):
		cmin = np.inf
		for j in range(nB):
			d = distance_function(XA[i,:], XB[j,:])
			if d<cmin:
				cmin = d
			if cmin<cmax:
				break
		if cmin>cmax and np.inf>cmin:
			cmax = cmin
	for j in range(nB):
		cmin = np.inf
		for i in range(nA):
			d = distance_function(XA[i,:], XB[j,:])
			if d<cmin:
				cmin = d
			if cmin<cmax:
				break
		if cmin>cmax and np.inf>cmin:
			cmax = cmin
	return cmax

def hausdorff_distance(XA, XB, distance='euclidean'):
	assert isinstance(XA, np.ndarray) and isinstance(XB, np.ndarray), \
		'arrays must be of type numpy.ndarray'
	assert np.issubdtype(XA.dtype, np.number) and np.issubdtype(XA.dtype, np.number), \
		'the arrays data type must be numeric'
	assert XA.ndim == 2 and XB.ndim == 2, \
		'arrays must be 2-dimensional'
	assert XA.shape[1] == XB.shape[1], \
		'arrays must have equal number of columns'
	
	if isinstance(distance, str):
		assert distance in _find_available_functions(distances), \
			'distance is not an implemented function'
		if distance == 'haversine':
			assert XA.shape[1] >= 2, 'haversine distance requires at least 2 coordinates per point (lat, lng)'
			assert XB.shape[1] >= 2, 'haversine distance requires at least 2 coordinates per point (lat, lng)'
		distance_function = getattr(distances, distance)
	elif callable(distance):
		distance_function = distance
	else:
		raise ValueError("Invalid input value for 'distance' parameter.")
	return _hausdorff(XA, XB, distance_function)



from distutils.core import setup
from setuptools import find_packages

setup( 
  name = "hausdorff",
  version = '0.2.6',
  author = 'mavillan',
  author_email = 'nallivam@gmail.com',
  packages=find_packages(),
  license='GNU GENERAL PUBLIC LICENSE Version 3',
  long_description="""
  This package implements the algorithm presented in An Efficient Algorithm for Calculating the Exact Hausdorff Distance (DOI: [10.1109/TPAMI.2015.2408351](https://doi.org/10.1109/TPAMI.2015.2408351)) by Aziz and Hanbury, 2015.
  """,
)



import pytest
import numba
import numpy as np
from hausdorff.hausdorff import hausdorff_distance
from scipy.spatial.distance import directed_hausdorff

@pytest.fixture
def input_arrays():
    np.random.seed(42)
    XA = np.random.random((1000,5))
    XB = np.random.random((2000,5))
    return (XA,XB)

def test_it_computes_haussdorf_distance(input_arrays):
    XA,XB = input_arrays

    actual_dist = hausdorff_distance(XA, XB, distance="euclidean")

    assert isinstance(actual_dist, float)

def test_it_correctly_computes_haussdorf_distance(input_arrays):
    XA,XB = input_arrays

    actual_dist = hausdorff_distance(XA, XB, distance="euclidean")
    expected_dist = max(directed_hausdorff(XA, XB)[0], directed_hausdorff(XB, XA)[0])

    np.testing.assert_almost_equal(expected_dist, actual_dist)

def test_it_computes_custom_haussdorf_distance(input_arrays):
    XA,XB = input_arrays
    @numba.jit(nopython=True, fastmath=True)
    def custom_dist(array_x, array_y):
        n = array_x.shape[0]
        ret = 0.
        for i in range(n):
            ret += (array_x[i]-array_y[i])**3 / (array_x[i]**2 + array_y[i]**2 + 0.1)
        return ret
        
    actual_dist = hausdorff_distance(XA, XB, distance=custom_dist)

    assert isinstance(actual_dist, float)



import pytest
import numpy as np
from hausdorff.hausdorff import hausdorff_distance

@pytest.fixture
def input_arrays():
    np.random.seed(42)
    XA = np.random.random((1000,5))
    XB = np.random.random((2000,5))
    return (XA,XB)

@pytest.mark.parametrize("distance", [
    "manhattan",
    "euclidean",
    "chebyshev",
    "cosine",
    "haversine",
])
def test_it_computes_haussdorf_distance(input_arrays, distance):
    XA,XB = input_arrays

    actual_dist = hausdorff_distance(XA, XB, distance=distance)

    assert isinstance(actual_dist, float)